if __debug__:
    from django.http import HttpRequest

import collections
from django_searchbar.forms import SearchBarForm
from django import forms
from django.middleware import csrf
from django.utils.safestring import mark_safe
from django.db.models import Q


def listify(item):
    """
    A simple function to create a list if item is not a list or a tuple
    @type item: str|iterable
    @return list
    """

    if not isinstance(item, (list, tuple)):
        item = [item]
    return item


class SearchBar(collections.MutableMapping):

    """
    Usage:
        In your view, do this:
        def my_view(request):
            search_bar = SearchBar(request, ['name', 'age'])
            if search_bar.is_valid():
                name_value = search_bar['name']
    """

    def __init__(self, request, fields=None, replacements={}, method='post'):

        assert isinstance(request, HttpRequest), 'request should be an instance of the HttpRequest object'
        assert isinstance(fields, (type(None), list, tuple, str, dict)), 'fields should be None, list or a tuple containing strings'
        assert isinstance(replacements, (dict, collections.Callable)), 'fields should be dictionary or a callable'

        if __debug__:
            def check_dict(item):
                assert 'label' in item, 'Your fields should have a label'
                if 'choices' in item:
                    assert isinstance(item['choices'], collections.Iterable), 'Your choices should be a dict'

            if isinstance(fields, (list, tuple)):
                for item in fields:
                    assert isinstance(item, (str, dict)), '%s should be a string or a dictionary containing label' % item
                    if isinstance(item, dict):
                        check_dict(item)
            elif isinstance(fields, dict):
                check_dict(fields)

        if fields:
            fields = listify(fields)

        self.request = request
        self.replacements = replacements
        self.fields = fields
        self.old_fields = None
        self.action = ''
        self.method = method.lower().strip()
        self.__form = None

    @property
    def form(self):
        if not self.__form and self.old_fields != self.fields:
            self.__form = SearchBarForm(self.request.GET or self.request.POST, fields=self.fields)
            self.__form.is_valid()
            self.old_fields = self.fields

        return self.__form

    def is_valid(self, *args, **kwargs):
        """
        Validates the SearchBar instance.
        All required argument you pass here, should end up in request results to pass.
        @return bool
        """

        def check_validation(self, item):
            if isinstance(item, dict):
                if item.get('required', False) and self.form.cleaned_data.get(item['label'], '') == '':
                    return False
            elif isinstance(item, str):
                if self.form.cleaned_data.get(item, '') == '':
                    return False

            return True

        if not self.fields:
            return False

        form_validation = self.form.is_valid()

        if form_validation and args:
            args = listify(args)
            form_validation = all(self.form.cleaned_data.get(item, '') != '' for item in args)

        elif form_validation and self.fields:
            for item in self.fields:
                if not check_validation(self, item):
                    form_validation = False
                    break

        return form_validation

    def as_form(self):
        csrf_ = ''
        if self.method == 'post':
            csrf_ = "<input type='hidden' name='csrfmiddlewaretoken' value='{0}' />".format(csrf.get_token(self.request))
        submit_button = '<input type="submit" value="submit" />'
        return_string = "<form method='%s' action='%s'>%s %s %s</form>" % (self.method, self.action, csrf_, self, submit_button)
        return mark_safe(return_string)

    def __getitem__(self, key):

        if key == 'as_form':
            return self.as_form()

        return self.form.cleaned_data.get(key, '')

    def get_filters(self, *args, lookup_string=''):
        """
        Returns a Q object based on all the input from query term
        @param lookup_string: adds this ``lookup_string`` to query lookup of all fields
        @param args: if provided, items you need to be in queryset. otherwise it's everything
        """
        filters = Q()
        lookup_string = lookup_string.lower().strip()

        if args:
            __fields = [k for k in self.fields if k in args]
        else:
            __fields = self.fields

        for field in __fields:
            if isinstance(field, dict):
                field = field['label']
            if self[field]:
                replacement = self.replacements.get(field, field)
                if isinstance(replacement, collections.Callable):
                    replacement = replacement(field)

                if lookup_string:
                    field_name = "{field}__{method}".format(field=replacement, method=lookup_string)
                else:
                    field_name = replacement

                filters &= Q(**{field_name: self[field]})
        return filters

    def __contains__(self, key):

        return key in self.form.fields

    def __setitem__(self, key, value):

        if isinstance(value, str):

            self.form.fields[key] = forms.CharField(label=value, required=False)

        elif isinstance(value, (list, tuple)):

            label = key.replace('-', ' ').replace('_', ' ').title()
            self.form.fields[key] = forms.ChoiceField(label=label, choices=value, required=False)

        elif isinstance(value, dict):

            required = value.get('required', False)

            if 'label' in value:
                label = value['label']
            else:
                label = key.replace('-', ' ').replace('_', ' ').title()

            self.form.fields[key] = forms.ChoiceField(label=label, choices=value, required=required)

    def __delitem__(self, key):

        self.form.fields.pop(key)

    def __iter__(self):

        return iter(self.form.fields)

    def __len__(self):

        return len(self.form.fields)

    def __str__(self):

        return str(self.form)




from django import template
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q
from people.models import Friendship

register = template.Library()


@register.filter(name='connection')
def connection(person, mate):
    try:
        friendship = Friendship.objects.get(from_person=person, to_person=mate)
        direction = 'left'
    except ObjectDoesNotExist:
        try:
            friendship = Friendship.objects.get(from_person=mate, to_person=person)
            direction = 'right'
        except ObjectDoesNotExist:
            return False

    return ({
        Friendship.APROVED: 'approved',
        Friendship.PENDING: 'pending',
        Friendship.BLOCKED: 'blocked',
    }.get(friendship.status), direction, friendship.approval_date or friendship.request_date or None)


@register.filter(name='requested_friendship')
def requested_friendship(person, mate):
    return person.requested_friendship(mate)


@register.filter(name='friendship_aproved')
def friendship_aproved(person, mate):
    return person.friendship_aproved(mate)


@register.filter
def in_loop(mate, person):
    return person.in_loop(mate)


class AssignFriendsToNode(template.Node):

    def __init__(self, obj, context_var_name):
        self.obj = template.Variable(obj)
        self.context_var_name = context_var_name

    def render(self, context):
        actor = self.obj.resolve(context)
        context[self.context_var_name] = Friendship.objects.filter(Q(to_person=actor) | Q(from_person=actor)).filter(status=Friendship.APROVED)
        return ''


@register.tag
def get_friends(parser, token):
    """
    Assigns the friends of the actor to a variable in the context
    {% get_friends actor as friend_list %}
    """
    try:
        bits = token.split_contents()
    except ValueError:
        raise template.TemplateSyntaxError('This tag requires arguments')

    if len(bits) != 4:
        raise template.TemplateSyntaxError('This tag requires exactly three arguments')
    if bits[2] != 'as':
        raise template.TemplateSyntaxError("Second argument to tag must be 'as'")

    return AssignFriendsToNode(bits[1], bits[3])



from django import forms


class SearchBarForm(forms.Form):

    """
    This is the form that ends up in your templates
    """

    def __init__(self, data, fields):

        from django_searchbar.utils import listify
        fields = listify(fields)

        if __debug__:
            for field in fields:
                assert isinstance(field, (str, dict)), 'fields should be string types, but you provided %s' % type(field)
                if isinstance(field, dict):
                    assert 'label' in field, 'You should provide a label'
                    if 'choices' in field:
                        assert isinstance(field['choices'], (list, tuple)), 'You should provide a label'

        super(SearchBarForm, self).__init__(data)
        for field in fields:

            if isinstance(field, str):

                label = field.replace('-', ' ').replace('_', ' ').title()
                self.fields[field] = forms.CharField(label=label, required=False)

            elif isinstance(field, dict):

                label = field['label'].replace('-', ' ').replace('_', ' ').title()
                required = field.get('required', False)

                if 'choices' in field:
                    self.fields[field['label']] = forms.ChoiceField(label=label, choices=field['choices'], required=required)
                else:
                    self.fields[field['label']] = forms.CharField(label=label, required=required)



# -*- coding: utf-8 -*-
from .utils import SearchBar


class SearchBarViewMixin:

    """
    A Class Based View Mixin to use SearchBar
    """
    searchbar_method = 'post'

    def get_searchbar(self, request):
        return SearchBar(request, self.searchbar_fields, self.replacements, self.searchbar_method)

    def dispatch(self, request, *args, **kwargs):
        assert hasattr(self, 'searchbar_fields'), 'You should provide searchbar_fields attribute in your class'
        self.replacements = getattr(self, 'searchbar_replacements', {})
        self.searchbar_obj = self.get_searchbar(request)
        return super().dispatch(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):

        search_obj = self.get_searchbar(request)
        if search_obj.is_valid():
            self.queryset = self.queryset.filter(search_obj.get_filters())

        if hasattr(super(), 'post'):
            return super().post(request, *args, **kwargs)
        else:
            return super().get(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):

        search_obj = self.get_searchbar(request)
        if search_obj.is_valid():
            self.queryset = self.queryset.filter(search_obj.get_filters())

        return super().get(request, *args, **kwargs)



from django.test import TestCase
from django.test.client import Client
from django.test.client import RequestFactory
from django.forms import fields
from django.db.models import Q
from django_searchbar.utils import SearchBar, listify
from django_searchbar.forms import SearchBarForm


class UtilsTestCase(TestCase):

    def testListify(self):
        self.assertIsInstance(listify('name'), (list, tuple))
        self.assertIsInstance(listify(['name']), (list, tuple))
        self.assertNotIsInstance(listify('name'), str)

    def testInitializationFailsWithWrongTypes(self):

        request = RequestFactory().get('/')
        self.assertRaises(TypeError, SearchBar)
        self.assertRaises(AssertionError, SearchBar, 1)
        self.assertRaises(AssertionError, SearchBar, request, [[]])
        self.assertRaises(AssertionError, SearchBar, request, [None])
        self.assertRaises(AssertionError, SearchBar, ['name'], request)
        self.assertIsInstance(SearchBar(fields=['name'], request=request), SearchBar)
        self.assertIsInstance(SearchBar(request, ['name']), SearchBar)
        self.assertIsInstance(SearchBar(request, 'name'), SearchBar)

    def testInitializationWithChoices(self):
        request = RequestFactory().get('/')
        self.assertIsInstance(SearchBar(request, {'label': 'name'}), SearchBar)
        self.assertIsInstance(SearchBar(request, {
            'label': 'name',
            'choices': ({'aaa': 'bbb'},),
        }), SearchBar)
        self.assertIsInstance(SearchBar(request, {
            'label': 'name',
            'choices': (('aaa', 'bbb'),),
        }), SearchBar)
        self.assertRaises((AssertionError, ValueError), SearchBar, request, {
            'labels': 'name',
            'choices': (('aaa', 'bbb'),),
        })
        self.assertRaises((AssertionError, ValueError), SearchBar, request, {
            'choices': (('aaa', 'bbb'),),
        })
        self.assertRaises((AssertionError, ValueError), SearchBar, request, {
            'label': 'name',
            'choices': None,
        })

    def testChoicesReturnValues(self):
        request = RequestFactory().get('/?name=arsham')
        search_bar = SearchBar(request, {
            'label': 'name',
            'choices': (
                ('arsham', 'Arsham Shirvani'),
                {'ivan': 'Ivan'},
            ),
        })
        search_bar.is_valid()
        self.assertEqual(search_bar['name'], 'arsham')
        self.assertEqual(search_bar.get('name'), 'arsham')

        request = RequestFactory().get('/?name2=arsham')
        search_bar = SearchBar(request, {
            'label': 'name',
            'choices': (
                ('arsham', 'Arsham Shirvani'),
                {'ivan': 'Ivan'},
            ),
        })
        search_bar.is_valid()
        self.assertEqual(search_bar['name'], '')

    def testRequiredFields(self):
        request = RequestFactory().get('/?name2=arsham')
        search_bar = SearchBar(request, {
            'label': 'name',
            'choices': (
                ('arsham', 'Arsham Shirvani'),
            ),
            'required': False,
        })

        self.assertTrue(search_bar.is_valid())
        self.assertFalse(search_bar.form.fields['name'].required)

        request = RequestFactory().get('/?name2=arsham')
        search_bar = SearchBar(request, {
            'label': 'name',
            'choices': (
                ('arsham', 'Arsham Shirvani'),
            ),
            'required': True,
        })

        self.assertFalse(search_bar.is_valid())
        self.assertTrue(search_bar.form.fields['name'].required)

    def testIsValidMethodGet(self):
        request = RequestFactory().get('/?name=123')
        search_bar = SearchBar(request, 'name')
        self.assertTrue(search_bar.is_valid())
        self.assertTrue(search_bar.is_valid('name'))

        request = RequestFactory().get('/?name=123&age=6&error=info')
        search_bar = SearchBar(request, ['name', 'age'])
        self.assertTrue(search_bar.is_valid())
        self.assertTrue(search_bar.is_valid('name'))
        self.assertTrue(search_bar.is_valid('age'))
        self.assertTrue(search_bar.is_valid('age', 'name'))

        request = RequestFactory().get('/?name=123')
        search_bar = SearchBar(request, ['name', 'age'])
        self.assertTrue(search_bar.is_valid('name'))

    def testIsValidMethodPost(self):
        request = RequestFactory().post('/', {'name': 123})
        search_bar = SearchBar(request, 'name')
        self.assertTrue(search_bar.is_valid())
        self.assertTrue(search_bar.is_valid('name'))

        request = RequestFactory().post('/', {'name': 123, 'age': 6, 'error': 'info'})
        search_bar = SearchBar(request, ['name', 'age'])
        self.assertTrue(search_bar.is_valid())
        self.assertTrue(search_bar.is_valid('name'))
        self.assertTrue(search_bar.is_valid('age'))
        self.assertTrue(search_bar.is_valid('age', 'name'))

        request = RequestFactory().post('/', {'name': 123})
        search_bar = SearchBar(request, ['name', 'age'])
        self.assertTrue(search_bar.is_valid('name'))

    def testIsValidMethodFails(self):
        request = RequestFactory().get('/')
        search_bar = SearchBar(request, 'name')
        self.assertFalse(search_bar.is_valid())
        self.assertFalse(search_bar.is_valid('name'))

        request = RequestFactory().get('/?name=123')
        search_bar = SearchBar(request, ['name', 'age'])
        self.assertFalse(search_bar.is_valid())
        self.assertFalse(search_bar.is_valid('age'))
        self.assertFalse(search_bar.is_valid('age', 'name'))

    def testGettingBackTheValue(self):
        request = RequestFactory().get('/?name=arsham&age=6')
        search_bar = SearchBar(request, ['name', 'age'])
        search_bar.is_valid()
        self.assertEqual(search_bar['name'], 'arsham')
        self.assertEqual(search_bar['age'], '6')

        request = RequestFactory().get('/?name=&age=')
        search_bar = SearchBar(request, ['name', 'age'])
        search_bar.is_valid()
        self.assertEqual(search_bar['name'], '')
        self.assertEqual(search_bar['age'], '')

    def testOutputString(self):
        request = RequestFactory().get('/')
        search_bar = SearchBar(request, ['name', 'age'])
        self.assertIn('id="id_name"', str(search_bar))
        self.assertIn('name="name"', str(search_bar))

        self.assertIn('csrfmiddlewaretoken', str(search_bar.as_form()))
        self.assertIn('<form', str(search_bar.as_form()))
        self.assertIn('</form>', str(search_bar.as_form()))

    def testGetFilters(self):
        request = RequestFactory().get('/?username=arsham&email=this_is_an_email')
        search_bar = SearchBar(request, ['username', 'email'])
        search_bar.is_valid()
        self.assertIsInstance(search_bar.get_filters(), Q)
        self.assertIn("'username'", str(search_bar.get_filters()))
        self.assertNotIn("'username__iexact'", str(search_bar.get_filters()))
        self.assertIn("'email'", str(search_bar.get_filters()))
        self.assertIn("'username__icontains'", str(search_bar.get_filters(lookup_string='icontains')))
        self.assertIn("'email__icontains'", str(search_bar.get_filters(lookup_string='icontains')))
        self.assertIn("(AND: ", str(search_bar.get_filters()))
        self.assertIn("arsham", str(search_bar.get_filters()))
        self.assertIn("this_is_an_email", str(search_bar.get_filters()))
        self.assertNotIn('(OR:', str(search_bar.get_filters()))

        search_bar = SearchBar(request, ['username', 'email'])
        search_bar.is_valid()
        self.assertIn("'username'", str(search_bar.get_filters('username')))
        self.assertNotIn("'email__exact'", str(search_bar.get_filters('username')))

        search_bar = SearchBar(request, ['username', 'email', 'age'])
        search_bar.is_valid()
        self.assertIn("'username'", str(search_bar.get_filters('username', 'age')))
        self.assertIn("'username__icontains'", str(search_bar.get_filters('username', 'age', lookup_string='icontains')))
        self.assertIn("'email'", str(search_bar.get_filters('username', 'email')))
        self.assertIn("'username'", str(search_bar.get_filters('username', 'email')))
        self.assertNotIn("'age'", str(search_bar.get_filters('username', 'email')))

    def testEmptyFilters(self):
        request = RequestFactory().get('/')
        search_bar = SearchBar(request, ['username', 'email'])
        search_bar.is_valid()
        self.assertEqual("(AND: )", str(search_bar.get_filters()))

    def testReplacementsDictionary(self):
        request = RequestFactory().get('/?name=arsham&age=6')
        search_bar = SearchBar(request, ['name'], replacements={'name': 'username'})
        search_bar.is_valid()
        self.assertIn("'username'", str(search_bar.get_filters()))
        self.assertNotIn("'user__'", str(search_bar.get_filters()))

    def testReplacementsCallable(self):
        request = RequestFactory().get('/?name=arsham&age=6')
        search_bar = SearchBar(request, ['name'], replacements={'name': lambda x: 'username'})
        search_bar.is_valid()
        self.assertIn("'username'", str(search_bar.get_filters()))
        self.assertNotIn("'user__'", str(search_bar.get_filters()))


    def testAddFields(self):

        request = RequestFactory().get('/?name=asdasd')
        search_bar = SearchBar(request, ['name'])

        search_bar.is_valid()

        self.assertNotIn('asd1', search_bar)
        search_bar['asd1'] = 'asd'
        self.assertIn('asd1', search_bar)

        search_bar['asd2'] = ({'aaa': 'bbb'},)
        self.assertIn('asd2', search_bar)

        search_bar['asd3'] = {'label': 'ASD', 'choices': ({'aaa': 'bbb'},)}
        self.assertIn('asd3', search_bar)


    def testPopFields(self):

        request = RequestFactory().get('/?name=asdasd&age=10&other=asdasd')
        search_bar = SearchBar(request, ['name', 'age', 'other'])

        search_bar.is_valid()

        self.assertNotIn('asd', search_bar)

        search_bar.pop('age')
        self.assertNotIn('age', search_bar)

        search_bar.pop('other')
        self.assertNotIn('other', search_bar)

        self.assertNotIn('ggg', search_bar)


class FormsTestCase(TestCase):

    def testInitializationFailsWithWrongTypes(self):
        self.assertRaises((AssertionError, TypeError), SearchBarForm)
        self.assertRaises((AssertionError, TypeError), SearchBarForm, 6)
        self.assertRaises((AssertionError, TypeError), SearchBarForm, ['666', 6])

    def testStuffThatAreNotInFormShouldFail(self):
        form = SearchBarForm({}, ['name', 'age'])
        self.assertIn('name', form.fields)
        self.assertIn('age', form.fields)
        self.assertNotIn('blah', form.fields)

    def testFormOutput(self):
        form = SearchBarForm({}, ['name', 'age'])
        self.assertIn('id_age', str(form))
        self.assertIn('id_name', str(form))

    def testInstantiateWithChoices(self):

        self.assertIsInstance(SearchBarForm({}, {
            'label': 'name',
            'choices': ({'aaa': 'bbb'},),
        }), SearchBarForm)
        self.assertRaises((AssertionError, ValueError), SearchBarForm, {}, {
            'labels': 'name',
            'choices': (('aaa', 'bbb'),),
        })
        self.assertRaises((AssertionError, ValueError), SearchBarForm, {}, {
            'choices': (('aaa', 'bbb'),),
        })
        self.assertRaises((AssertionError, ValueError), SearchBarForm, {}, {
            'label': 'name',
            'choices': None,
        })

    def testChoicesFieldHasADifferentWidget(self):

        search_bar = SearchBarForm({}, ['username', {
            'label': 'name',
            'choices': ({'aaa': 'bbb'},),
        }])
        self.assertIsInstance(search_bar.fields['username'], fields.CharField)
        self.assertIsInstance(search_bar.fields['name'], fields.ChoiceField)


class IntegrationTestCase(TestCase):

    def setUp(self):

        self.c = Client()

    def testCreatingFormWidgetOnOutput(self):
        response = self.c.get('/?name=arsham&age=6&order_by=asc')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.templates[0].name,
                         'django_searchbar/test.html')

        form = response.context['search_bar']
        self.assertIn('<label for=', str(form))
        self.assertIn('id="id_name"', str(form))
        self.assertIn('name="name"', str(form))
        self.assertIn('type="text"', str(form))
        self.assertIn('value="arsham"', str(form))
        self.assertIn('<select id="id_order_by"', str(form))
        self.assertIn('name="order_by"', str(form))
        self.assertIn('<option value="asc"', str(form))
        self.assertIn('selected="selected"', str(form))
        self.assertEqual(str(form).count('<option'), 2)
        self.assertEqual(str(form).count('<select'), 1)
        self.assertEqual(str(form).count('</option>'), 2)

        form.is_valid()

        self.assertEqual(form['name'], 'arsham')
        self.assertEqual(form['age'], '6')
        self.assertEqual(form['order_by'], 'asc')



import os
BASE_DIR = os.path.dirname(os.path.dirname(__file__))
SECRET_KEY = 'kjuiyhgwkjfygwruig389o7tgo87gfheirgo3e8574o3g8i4gfui'
DEBUG = True
TEMPLATE_DEBUG = True
ALLOWED_HOSTS = []

INSTALLED_APPS = (
    'django.contrib.sessions',
    'django.contrib.messages',
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
)

ROOT_URLCONF = 'django_searchbar.urls'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

LANGUAGE_CODE = 'en-gb'
TIME_ZONE = 'UTC'
USE_I18N = False
USE_L10N = False
USE_TZ = True

STATIC_URL = '/static/'

TEMPLATE_DIRS = (
    os.path.join(BASE_DIR, 'django_searchbar/templates'),
)

PASSWORD_HASHERS = (
    'django.contrib.auth.hashers.MD5PasswordHasher',
)



# -*- coding: utf-8 -*-

import os
from setuptools import setup, find_packages


def read_file(filename):
    return open(os.path.join(os.path.dirname(__file__), filename)).read()


setup(
    name='django-searchbar',
    version=__import__('django_searchbar').__version__,
    description="Simple searchbar and handler you can use in all your views and templates.",
    long_description=read_file('README.md'),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Software Development :: Libraries',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Framework :: Django',
        'Environment :: Web Environment',
    ],
    keywords=['searchbar', 'django', 'forms', 'template'],
    platforms=['OS Independen'],
    author='Arsham Shirvani',
    author_email='arshamshirvani@gmail.com',
    url='http://github.com/arsham/django-searchbar',
    license='BSD',
    packages=find_packages(),
    zip_safe=False,
)



from django.shortcuts import render
from django_searchbar.utils import SearchBar


def homepage(request):

    search_bar = SearchBar(request, [
        'name', {
            'label': 'age',
        },
        {
            'label': 'order_by',
            'choices': (
                ('asc', 'ASC'),
                ('desc', 'DESC'),
            ),
        }
    ])

    return render(request, 'django_searchbar/test.html', {
        'search_bar': search_bar,
    })
